/*
 * pawstextwrap.cpp
 *
 * Copyright (C) 2006 Atomic Blue (info@planeshift.it, http://www.atomicblue.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
#include <psconfig.h>
#include <ctype.h>
#include <ivideo/fontserv.h>
#include <iutil/virtclk.h>
#include <iutil/evdefs.h>
#include <iutil/event.h>

#include "pawstextbox.h"
#include "pawsmanager.h"
#include "pawsprefmanager.h"
#include "pawscrollbar.h"
#include "util/log.h"
#include "util/strutil.h"
#include "util/psstring.h"

#define EDIT_TEXTBOX_MOUSE_SCROLL_AMOUNT 3

#define VSCROLLBAR_WIDTH 24

/* 
   In order for this control to be implemented correctly, it must pass the following tests:
     1. Takes the last word after a space on the line and moves it to the next line. (wordwrap)
     2. Chops the end of the line and continues on the next line if no space is found. (breakline)
     3. The line separator(OS dependant - \n for our purposes) has adjustments made for it. (line separator)
     4. Combinations of the above must also be considered, such as: 
           a. After a breakline occurs, the text will still wrap to the next line.
           b. Before a breakline occurs, the line starts from being wordwrapped on the previous line.
           c. A newline must work from either a line containing a wordwrap or a breakline.
     5. Multiple blank lines must be allowed.
     6. An control with no text must be considered.
     7. Cursor positioning must be considered for all of the above tests.
     8. The vertical scrollbar must be considered for all of the above tests.
     9. Delete, Backspace and insert text must work in the beginning, middle, 
          and end of the text.
    10. Font support by control
*/
pawsMultilineEditTextBox::pawsMultilineEditTextBox()
{
    blink = true;
    cursorPosition = 0;
    cursorLine = 0;
    topLine=0;
    vScrollBar = NULL;
    
    clock =  csQueryRegistry<iVirtualClock > ( PawsManager::GetSingleton().GetObjectRegistry());

    blinkTicks = clock->GetCurrentTicks();

    // Find the line height;
    int dummy;
    GetFont()->GetMaxSize( dummy, lineHeight );
    lineHeight -=2;
    OnResize();
}

pawsMultilineEditTextBox::~pawsMultilineEditTextBox()
{

}


bool pawsMultilineEditTextBox::OnKeyDown( int code, int key, int modifiers )
{   
    bool changed = false;
    size_t position = 0;
    bool repositionCursor = false;

    blink = true;

    switch ( key )
    {
    case CSKEY_DEL:
    {
        size_t position =  GetCursorPosition(cursorLine,cursorLoc);
        if (text.Length() >0 && position < text.Length())
        {
            text.DeleteAt(position);
            changed = true;
        }
        break;
    }
    case CSKEY_BACKSPACE:
    {
         position = GetCursorPosition(cursorLine, cursorLoc);
        if (text.Length() > 0 && position >0 )//&& position <=text.Length())
        {
            position--;        
            text.DeleteAt(position);
            repositionCursor = true;
            changed = true;
        }
        break;
    }
    case CSKEY_LEFT:
        if ( cursorLoc > 0 )
            cursorLoc--;
        else if (cursorLine > 0)
        {
            cursorLine--;
            cursorLoc = lineInfo[cursorLine]->lineLength -1 ;
            break;
        }
        break;
    case CSKEY_RIGHT:
        if ( cursorLoc < lineInfo[cursorLine]->lineLength-lineInfo[cursorLine]->lineExtra )
        {
            cursorLoc++;
        }
        else if (cursorLine<lineInfo.GetSize()-1)
        {
            cursorLine++;
            cursorLoc=0;
        } else if (cursorLine == lineInfo.GetSize()-1 && cursorLoc == lineInfo[cursorLine]->lineLength-1) {
          //the case for the very last line, we allow to go to after the last character
           cursorLoc++; 
        }
        break;
    case CSKEY_UP:
    {
        if (cursorLine>0)
            cursorLine--;
        if(cursorLoc > lineInfo[cursorLine]->lineLength - lineInfo[cursorLine]->lineExtra)
            cursorLoc = lineInfo[cursorLine]->lineLength - lineInfo[cursorLine]->lineExtra;
        break;
    }
    case CSKEY_DOWN:
    {
        if (cursorLine < lineInfo.GetSize()-1)
            cursorLine++;
        if(cursorLoc > lineInfo[cursorLine]->lineLength - lineInfo[cursorLine]->lineExtra)
            cursorLoc = lineInfo[cursorLine]->lineLength - lineInfo[cursorLine]->lineExtra;
        break;
    }
    case CSKEY_END:
    {
        cursorLoc = lineInfo[cursorLine]->lineLength-lineInfo[cursorLine]->lineExtra;
        break;
    }
    case CSKEY_HOME:
    {
        cursorLoc = 0;
        break;
    }
    default:
        if ( !isprint((unsigned char)key) && !(key==CSKEY_ENTER))
        {
            //unhandled non-printing character
            break;
        }            

        if ( key == CSKEY_ALT || key == CSKEY_CTRL )
        {
            break;
        }
        //Lookup the position.
        position =  GetCursorPosition(cursorLine,cursorLoc);
        if (key == CSKEY_ENTER)
        {
            //our text flow algorithm will then break this into two lines.
            text.Insert( position, '\n');
            cursorLine++;
            cursorLoc = 0;
        }
        else
        {
            text.Insert( position, (char) key);
            repositionCursor = true;
            position++;
        }
        
        changed = true;
    }

    if (changed)
    {
        //Re-flow the text
        LayoutText();
        
        if (repositionCursor)
           GetCursorLocation(position,cursorLine,cursorLoc);
 
        if (subscribedVar)
            PawsManager::GetSingleton().Publish(subscribedVar, text);
        parent->OnChange(this);
    }

    if (cursorLine > topLine+canDrawLines)
        topLine = cursorLine - canDrawLines;
    else if (cursorLine < topLine)
        topLine = cursorLine;
    SetupScrollBar();

    if ( isprint((unsigned char)key)  )       
        return true;
    
    if (!(code== CSKEY_ENTER))
    {        
        pawsWidget::OnKeyDown( code, key, modifiers );
    }

    return true;
}

void pawsMultilineEditTextBox::OnResize()
{
    canDrawLines = (screenFrame.Height()-lineHeight) / lineHeight;
}

void pawsMultilineEditTextBox::SetupScrollBar()
{
    if (!vScrollBar)
    {
        vScrollBar = new pawsScrollBar;
        vScrollBar->SetParent( this );
        vScrollBar->SetRelativeFrame( screenFrame.Width() - VSCROLLBAR_WIDTH,
                                      0, VSCROLLBAR_WIDTH, screenFrame.Height() - 12 );
        vScrollBar->SetAttachFlags(ATTACH_TOP | ATTACH_BOTTOM | ATTACH_RIGHT);
        vScrollBar->PostSetup();
        vScrollBar->SetTickValue( 1.0 );
        AddChild( vScrollBar ); 
    }

    if (lineInfo.GetSize() < canDrawLines)
    {
        // turn off scrollbar
        vScrollBar->Hide();
    } 
    else 
    {
        vScrollBar->SetCurrentValue((float)topLine);
        vScrollBar->SetMaxValue((int)lineInfo.GetSize() - canDrawLines );
        vScrollBar->Show();
    }
}

bool pawsMultilineEditTextBox::OnMouseDown( int button, int modifiers, int x, int y )
{
    if (button == csmbLeft)
    { 
        return true;
    }
    else if (button == csmbWheelUp)
    {
        if (vScrollBar)
            vScrollBar->SetCurrentValue(vScrollBar->GetCurrentValue() - EDIT_TEXTBOX_MOUSE_SCROLL_AMOUNT);
        return true;
    }
    else if (button == csmbWheelDown)
    {
        if (vScrollBar)
            vScrollBar->SetCurrentValue(vScrollBar->GetCurrentValue() + EDIT_TEXTBOX_MOUSE_SCROLL_AMOUNT);
        return true;
    } 
    return pawsWidget::OnMouseDown( button, modifiers, x ,y);
}

bool pawsMultilineEditTextBox::OnMouseUp( int button, int modifiers, int x, int y )
{
    if (button == csmbLeft)
    {
        size_t EndLine = 0;
        size_t EndChar = 0;
        CalcMouseClick(x,y,EndLine, EndChar);
        
        //Set the cursor location.
        cursorLine = (unsigned int) EndLine;
        cursorLoc = EndChar;
        cursorPosition = GetCursorPosition();

        return true;
    } 
    return pawsWidget::OnMouseUp( button, modifiers, x ,y);
}

void pawsMultilineEditTextBox::CalcMouseClick( int x, int y, size_t &cursorLine, size_t &cursorChar)
{
    // Adjust x to be relative to the text box
    x -= (screenFrame.xmin+4); 

    //Force the cursor to blink
    blink = true;

    // Find the line the mouse is on and move the cursor there
    cursorLine=(y-screenFrame.ymin)/lineHeight+topLine;
    
    //Determine if the click was after the text area
    if (cursorLine >= lineInfo.GetSize() && lineInfo.GetSize() >0)
    {
        cursorLine = (unsigned int) lineInfo.GetSize()-1;
        //*after* the last letter of last line
        cursorChar = lineInfo[(int)lineInfo.GetSize()-1]->lineLength;
        return;
    }
    // Determine if the click was before text area
    if (x<=0 || lineInfo[cursorLine]->lineLength <=0)
    {
        cursorChar=0;
        return;
    }
 
    //Get line width
    int width = 0;
    int height = 0;
    csString line = GetLine(cursorLine);
    GetFont()->GetDimensions(line,width,height);
    //Determine if the click was at the end of a line...
     if ((x>=width)) 
    {
        if (line.GetAt(lineInfo[cursorLine]->lineLength-1) == '\n') 
        {
            //if it was, make sure to ignore the newline character.
            cursorChar = lineInfo[cursorLine]->lineLength-1;
            return;
        }
    }
    //should position cursor before clicked letter
    cursorChar = GetFont()->GetLength(GetLine(cursorLine),x);
}


void pawsMultilineEditTextBox::OnUpdateData(const char *dataname,PAWSData& value)
{
    // This is called automatically whenever subscribed data is published.
    SetText( value.GetStr(), false );
}

bool pawsMultilineEditTextBox::OnScroll( int direction, pawsScrollBar* widget )
{    
    topLine = (int)widget->GetCurrentValue();
    return true;
}
void pawsMultilineEditTextBox::PushLineInfo(size_t lineLength, size_t lineBreak, int lineExtra)
{
    MessageLine* msg = new MessageLine;
    msg->lineLength = lineLength;
    msg->breakLine = lineBreak;
    msg->lineExtra = lineExtra;
    lineInfo.Push(msg);
}

void pawsMultilineEditTextBox::LayoutText()
{
    csRef<iFont> font = GetFont();
    int screenWidth = screenFrame.Width()-VSCROLLBAR_WIDTH;
    int width = 0;
    int height = 0;
    csString tempString;
    csString word;
    size_t  srcPos = 0; //the first char that hasn't yet been grabbed.
    lineInfo.Empty();
    size_t totalCount = 0;
    int tailWidth = 0;
    
    while(srcPos < text.Length())
    {
        //try to grab a word.
        size_t a = text.FindFirst(" \t\n",srcPos);
        //if a is '-1', the char wasn't found so we grab to end of the line, 
        //else we grab up to and including the found character
        if (a == (size_t)-1)
            word = text.Slice(srcPos,text.Length()-srcPos);
        else 
            word = text.Slice(srcPos, a-srcPos+1);

        //Gets dimensions of current word.    
        font->GetDimensions(word,width,height);
        
        //if it'll fit..
        if(width+tailWidth <= screenWidth)
        {
            //then attach it
            tempString.Append(word);
            font->GetDimensions(tempString,tailWidth,height);
            srcPos += word.Length();
        } 
        else if(width > screenWidth) 
        {
            //Wordwrap 
            totalCount += tempString.Length();
            PushLineInfo(tempString.Length(),totalCount,1);
            tempString = word;
            font->GetDimensions(tempString,tailWidth,height);
            srcPos += word.Length();
        }
        else if (tempString.FindFirst(" ") != (size_t)-1)
        {
            //If wordwrap->linebreak on same line, 
            // push linebreak to next line
            totalCount += tempString.Length();
            PushLineInfo(tempString.Length(),totalCount,1);
            tempString.Clear();
            word = "a"; //set position holder (never actually drawn)
            font->GetDimensions(tempString,tailWidth,height);
            //Note: srcPos is not set.
        }
        else 
        {   
            //breakline - the case where we have a single 
            //unbroken word wider than our page

            int maxChars = font->GetLength(word,screenWidth); 
            word = word.Slice(0,maxChars);    //Get string for the line
            totalCount += word.Length();     //Get absolute char position in text for line break
            PushLineInfo(word.Length(),totalCount,0);//Push the line information to the stack
            srcPos += word.Length();//Set srcPosition for this loop
            //Note: Tempstring is not set here.    
            font->GetDimensions(tempString,tailWidth,height);
        }
        //does our word end in a newline?
        if(word[word.Length()-1] == '\n')
        {
            //standard 'create a new destination line' procedure
            totalCount += tempString.Length();
            PushLineInfo(tempString.Length(),totalCount,1);
            tempString.Clear();
            font->GetDimensions(tempString,tailWidth,height);
        }
    }

    //Make sure to add the last line if it exists.
    totalCount += tempString.Length();
    PushLineInfo(tempString.Length(),totalCount,0);
}

//our draw is based on static, wrapped widget
void pawsMultilineEditTextBox::Draw()
{
    yPos = 0;

    if ( clock->GetCurrentTicks() - blinkTicks > BLINK_TICKS )
    {
        blink = !blink;
        blinkTicks = clock->GetCurrentTicks();
    }

    pawsWidget::Draw();
    pawsWidget::ClipToParent();
    
    for (size_t line = topLine; line <= (topLine+canDrawLines); line++ )
    {
        if ( line >= lineInfo.GetSize())
            break;
        //chomp newline characters
        if (lineInfo[line]->lineLength >0) 
        {
            tmp = GetLine(line).Slice(0,lineInfo[line]->lineLength - lineInfo[line]->lineExtra);
            DrawWidgetText(tmp,(int) screenFrame.xmin, (int) (screenFrame.ymin+yPos*lineHeight),-1,GetFontColour());
            yPos++;
        }
     }

    if (blink && hasFocus) // Draw the cursor
    {
        int width, height;
        GetFont()->GetDimensions( GetLine(cursorLine).Slice(0,cursorLoc).GetDataSafe(), width, height );
        size_t realHeight=lineHeight*(cursorLine-topLine);

        graphics2D->DrawLine( (float)(screenFrame.xmin + width + 1),
            (float)(screenFrame.ymin + realHeight),
            (float)(screenFrame.xmin + width + 1),
            (float)(screenFrame.ymin + realHeight + lineHeight), 
            GetFontColour() );
    }

}

void pawsMultilineEditTextBox::DrawWidgetText(const char *text, size_t x, size_t y, int style, int fg)
{
    csRef<iFont> font = GetFont();
    if (style==-1)
        style = GetFontStyle();

    if (fg == -1)
        fg = GetFontColour();

    if (parent && !parent->GetBackground().IsEmpty() && parent->isFadeEnabled() && parent->GetMaxAlpha() != parent->GetMinAlpha())
    {
        int a = (int) (255 - (parent->GetMinAlpha() + (parent->GetMaxAlpha()-parent->GetMinAlpha()) * parent->GetFadeVal() * 0.010));
        int r, g, b;

        if (style & FONT_DROPSHADOW)
        {
            graphics2D->GetRGB(GetFontShadowColour(), r, g, b);
            graphics2D->Write( font, (int) x+2, (int) y+2, graphics2D->FindRGB(r, g, b, a), -1, text );
        }
        graphics2D->GetRGB(GetFontColour(), r, g, b);
        graphics2D->Write( font, (int) x, (int) y, graphics2D->FindRGB(r, g, b, a), -1, text);
    }
    else
    {
        if (style & FONT_DROPSHADOW)
            graphics2D->Write( font, (int) x+2, (int) y+2, GetFontShadowColour(), -1, text );

        graphics2D->Write( font, (int) x, (int) y, fg, -1, text);    
    }
}

bool pawsMultilineEditTextBox::SelfPopulate( iDocumentNode *node)
{
    if (node->GetAttributeValue("text"))
    {
        SetText(node->GetAttributeValue("text"));
        return true;
    }
    else
        return false;
}

bool pawsMultilineEditTextBox::Setup( iDocumentNode* node )
{    
    csRef<iDocumentNode> textNode = node->GetNode( "text" );        
    
    if ( textNode )
    {        
        csRef<iDocumentAttribute> textAttribute = textNode->GetAttribute("string");
        if ( textAttribute )
        {            
            SetText( textAttribute->GetValue() );
        }
    } else SetText("");             
    return true;
}

void pawsMultilineEditTextBox::SetText( const char* newText, bool publish )
{
    if (publish && subscribedVar)
        PawsManager::GetSingleton().Publish(subscribedVar, newText);

    cursorPosition = 0;
    cursorLine = 0;
    topLine = 0;

    text = newText;

    LayoutText();

    SetupScrollBar();
}


const char* pawsMultilineEditTextBox::GetText() 
{
    return text;
}

//Get start and endpoints for specified line in text
void pawsMultilineEditTextBox::GetLinePos(size_t lineNumber, size_t &start, size_t &end)
{
    start = 0;
    end = 0;
    for(size_t i = 0; i < lineNumber;i++)
    {
        size_t imp = lineInfo[i]->lineLength;
        start += imp;
    }
    end = start + lineInfo[lineNumber]->lineLength;
}

const char pawsMultilineEditTextBox::GetAt(size_t destLine, size_t destCursor)
{
    return GetLine(destLine).GetAt(destCursor);
}

void pawsMultilineEditTextBox::GetLineRelative(size_t pos, size_t &start, size_t &end)
{
    size_t count = 0;

    for (size_t i = 0; i < lineInfo.GetSize();i++)
    {
        count = lineInfo[i]->breakLine- lineInfo[i]->lineExtra;
        if (pos <= count)
        {
            start = lineInfo[i]->breakLine - lineInfo[i]->lineLength;
            end = lineInfo[i]->breakLine;
            return;
        }
    }
}

int pawsMultilineEditTextBox::GetLineWidth(int lineNumber)
{
    int width = 0;
    int height = 0;
    csString line = GetLine(cursorLine);
    GetFont()->GetDimensions(line,width,height);
    return width;
}

bool pawsMultilineEditTextBox::EndsWithNewLine(int lineNumber)
{
    return GetLine(cursorLine).GetAt(lineInfo[lineNumber]->lineLength-1) == '\n' ? true : false;
}

//Get specified line in text
csString pawsMultilineEditTextBox::GetLine(size_t lineNumber)
{
    size_t start = 0;
    size_t end = 0;
    GetLinePos(lineNumber, start, end);
    return text.Slice(start, end - start);
}

//Get cursor position relative to entire text.
size_t pawsMultilineEditTextBox::GetCursorPosition()
{
    size_t count = 0;
    for(size_t i = 0; i < cursorLine;i++)
    {
        size_t pos = lineInfo[i]->lineLength;
        count += pos;
    }
    count = count + cursorLoc;
    return count;
}

size_t pawsMultilineEditTextBox::GetCursorPosition(size_t destLine, size_t destCursor)
{
    if (destLine >0)
        return lineInfo[destLine-1]->breakLine+destCursor;
    else
        return destCursor;
}

void pawsMultilineEditTextBox::GetCursorLocation(size_t pos, size_t &destLine, size_t &destCursor)
{
    size_t start = 0;
    size_t end = 0;
    destLine = 0;
    destCursor = 0;

    GetLineRelative(pos, start, end);
    for (size_t i = 0; i < lineInfo.GetSize(); i++)
    {
       //Adjustment for wordwrap so final character is not exceeded by the cursor
       if ((pos > (lineInfo[i]->breakLine - lineInfo[i]->lineExtra)) && (i+1<lineInfo.GetSize()))
       {    
          destLine++;
       }
       else 
       {
           destCursor = pos - start;
           return;
       }
    }
}

